#include "Inheritance_Matrix.h"
#include <math.h>
#include <bitset>
#include <iostream>

using namespace std;

Inheritance_Matrix::Inheritance_Matrix(Pedigree *p)
{
  ped = p;
  Initialize();
}

Inheritance_Matrix::Inheritance_Matrix()
{
  Initialize();
}

Inheritance_Matrix::~Inheritance_Matrix()
{
  if(bit_checker!=NULL){ delete [] bit_checker; bit_checker=NULL; }
  if(bit_count!=NULL) { delete [] bit_count; bit_count = NULL; }
  if(founder_node_indicator!=NULL)
    {
      for(int i=0; i<pow(4, nFounders); i++)
	delete [] founder_node_indicator[i];
      delete [] founder_node_indicator;
      founder_node_indicator = NULL;
    }
  
  if(geno_index!=NULL)
    {
      for(int iv=0; iv<nInherit_vector; iv++)
	{
	  for(int i=0; i<pow(4, nFounders); i++)
	    delete [] geno_index[iv][i];
	  delete [] geno_index[iv];
	}
      delete [] geno_index;
      geno_index = NULL;
    }
  
  if(inherit_matrix!=NULL)
    {
      delete [] inherit_vector;
      for(int i=0; i<nInherit_vector;i++)
	delete [] inherit_matrix[i];
      delete [] inherit_matrix;
      inherit_matrix = NULL;
    }
  
}

void Inheritance_Matrix::Initialize()
{
  fam = NULL;
  famSize = 0;
  nFounders = 0;
  nNonFounders = 0;
  two_n = 0;
  two_n_minus_f = 0;
  two_f = 0;
  bit_checker = NULL;
  bit_count = NULL;
  founder_node_indicator = NULL;
  geno_index = NULL;
  nInherit_vector=0;
  inherit_vector = NULL;
  inherit_matrix = NULL;
}

void Inheritance_Matrix::SetPedigree(Pedigree *p) { ped = p; }

void Inheritance_Matrix::SetFamily(Family* f)
{
  fam = f;
  famSize = fam->count;
  nFounders = fam->founders;
  nNonFounders = famSize - nFounders;
  two_n = 2*nNonFounders;
  two_f = 2*nFounders;
  two_n_minus_f = two_n - nFounders;
  four_pow_f = int(pow(4, nFounders));
  nInherit_vector = int(pow(2, two_n_minus_f));
  fa_traverse.resize(famSize);
  mo_traverse.resize(famSize);
  bit_checker = new int[two_f];
  bit_count = new uint8_t[nInherit_vector];
  founder_node_indicator = new uint8_t*[int(pow(4, nFounders))];

  for(int i=0; i<pow(4, nFounders); i++)
    founder_node_indicator[i] = new uint8_t[two_f];

  // Set bitset for inheritance vector (represented by unsigned int) to quickly check a specific bit
  //inherit_vector_bitset.resize(nInherit_vector);
  //for(unsigned int i=0; i<nInherit_vector; i++) {
  // std::bitset<sizeof(unsigned int)> bits(i);
  // inherit_vector_bitset[i] = bits;
 // }
}

void Inheritance_Matrix::SetParentalTraverse()
{
  Person *p;
  
  for(int i=0; i<famSize; i++)
    {
      p = ped->persons[fam->path[i]];
      if(p->isFounder()) 
	{
	  fa_traverse[i] = -1;
	  mo_traverse[i] = -1;
	  continue;
	}
      
      fa_traverse[i] = p->father->traverse;
      mo_traverse[i] = p->mother->traverse;   
    }
}

void Inheritance_Matrix::SetZeroBits()
{
  std::map<int, int> founder_first_offspring;
  zero_bits_index.clear();
  
  Person *p;
  for(int i=nFounders; i<famSize; i++)
    {
      p = ped->persons[fam->path[i]];
      fa_traverse[i] = p->father->traverse;
      mo_traverse[i] = p->mother->traverse;
      
      if(p->father->isFounder() && founder_first_offspring[fa_traverse[i]]==0)
	{
	  zero_bits_index.push_back(2*i);
	  founder_first_offspring[fa_traverse[i]] = i;
	}
      if(p->mother->isFounder() && founder_first_offspring[mo_traverse[i]]==0)
	{
	  zero_bits_index.push_back(2*i+1);
	  founder_first_offspring[mo_traverse[i]] = i;
	}
    }
  
  zero_bits_checker = 0;
  for(int i=0; i<nFounders; i++) {
    zero_bits_checker += 1<<(2*famSize-1-zero_bits_index[i]);
  }
  //cout <<"zero_bits_checker: "<<bitset<20>(zero_bits_checker).to_string() << endl;
}

void Inheritance_Matrix::Find_Founder_Node(int iv, uint8_t* iv_founder)
{
  for(int i=0; i<two_f; i++)
    iv_founder[i] = i;
  
  for(int i=two_f; i<2*famSize; i++)
    iv_founder[i] = Find_Founder_Node(iv, i);
}

//Find the founder node index for each bit indicagted by index in inheritance vector iv.
int Inheritance_Matrix::Find_Founder_Node(int iv, int index)
{
  int parent_traverse;
  int grand_paternal_node_index;
  int pos;

  int person_index = int(floor(index/2));
  int parental_index = index%2;

  if(parental_index==0) parent_traverse = fa_traverse[person_index];
  if(parental_index==1) parent_traverse = mo_traverse[person_index];

  pos = 2*famSize-1-index;
  int bit_index = 1<<pos;

  grand_paternal_node_index = (iv & bit_index)==0 ? parent_traverse*2 : parent_traverse*2+1;

  if(grand_paternal_node_index>=two_f) return Find_Founder_Node(iv, grand_paternal_node_index);
  
 //   printf("Pidex: %d partentIdx: %d IV: %d Index: %d Pos: %d and bit_index: %d nInherit_vec: %d ivANDbitIndex: %d grand_paternale_index: %d\n", 
 // person_index, parental_index, iv, index, pos, bit_index, nInherit_vector, iv&bit_index, grand_paternal_node_index);
  
  return(grand_paternal_node_index);
}

void Inheritance_Matrix::SetFounderNode()
{
  for(int i=0; i<two_f; i++)
    bit_checker[i] = 1<<(two_f-1-i);
  
  for(int i=0; i<pow(4, nFounders); i++)
    for(int j=0; j<two_f; j++)
      founder_node_indicator[i][j] = (i & bit_checker[j]) == 0 ? 0 : 1;
}

//Map founder node to 0 or 1 for each of 4^f combination
//Each of the combinations is represented by a number from 0 to 4^f-1
void Inheritance_Matrix::MapFounderNode()
{
  geno_index = new uint8_t **[nInherit_vector];
  for(int iv=0; iv<nInherit_vector; iv++)
    {  
      geno_index[iv] = new uint8_t*[int(pow(4, nFounders))];
      for(int i=0; i<pow(4, nFounders); i++)
	geno_index[iv][i] = new uint8_t[famSize];
    }
  
  for(int i=0; i<two_f; i++)
    bit_checker[i] = 1<<(two_f-1-i);

  for(int i=0; i<pow(4, nFounders); i++)
    for(int j=0; j<two_f; j++)
      founder_node_indicator[i][j] = (i & bit_checker[j]) == 0 ? 0 : 1;

  int pre;
#pragma omp parallel for private(pre)
  for(int iv=0; iv<nInherit_vector; iv++)
    for(int i=0; i<pow(4, nFounders); i++)
      for(int j=0; j<2*famSize; j++)
	{ 
	  if(j%2==1)
	    geno_index[iv][i][(j-1)/2] = pre + founder_node_indicator[i][inherit_matrix[iv][j]];
	  pre = founder_node_indicator[i][inherit_matrix[iv][j]];
	} 
}

void Inheritance_Matrix::CalcFounderGenoMappingByIV(int iv, std::vector<std::vector<uint8_t> >& geno_index_onfly)
{
    int pre;
    for(int i=0; i<pow(4, nFounders); i++)
      for(int j=0; j<2*famSize; j++)
        { 
          if(j%2==1)
            geno_index_onfly[i][(j-1)/2] = pre + founder_node_indicator[i][inherit_matrix[iv][j]];
          pre = founder_node_indicator[i][inherit_matrix[iv][j]];
        } 
}

void Inheritance_Matrix::Create_Inheritance_Matrix()
{
  SetParentalTraverse();
  SetZeroBits();

  inherit_vector = new unsigned int[nInherit_vector];
  inherit_matrix = new uint8_t*[nInherit_vector];

#pragma omp parallel for
  for(int i=0; i<nInherit_vector;i++)
    {
      inherit_matrix[i] = new uint8_t[2*famSize];
      inherit_vector[i] = 0;
    }

  int counter = 0;
  for(unsigned int i=0; i<pow(2, 2*nNonFounders); i++)
    {
      if((i & zero_bits_checker)>0) continue;
      inherit_vector[counter] =  i;
      
      bit_count[counter] = Count_Bits(i);
      
      Find_Founder_Node(inherit_vector[counter], inherit_matrix[counter]);    
      
      counter++;
    }
}

// Needed for the naive implementation of forward/backford computation or Viterbi algorith.
void Inheritance_Matrix::Count_Trans_Bits()
{
  for(unsigned int i=0; i<nInherit_vector; i++)
    for(unsigned int j=0; j<nInherit_vector; j++)
      bit_count_trans[inherit_vector[i]^inherit_vector[j]] = Count_Bits(inherit_vector[i]^inherit_vector[j]);
}

/*
  int Inheritance_Matrix::Count_Bits(unsigned int v)
  {
  int c;
  v = v - ((v >> 1) & 0x55555555);                    // reuse input as temporary
  v = (v & 0x33333333) + ((v >> 2) & 0x33333333);     // temp
  c = ((v + (v >> 4) & 0xF0F0F0F) * 0x1010101) >> 24; // count
  
  return c;
  }
*/

int Inheritance_Matrix::Count_Bits(unsigned int u)
{
  unsigned int uCount;
  
  uCount = u
    - ((u >> 1) & 033333333333)
    - ((u >> 2) & 011111111111);
  
  int c =   ((uCount + (uCount >> 3))
	     & 030707070707) % 63;
  
  // printf("UNIT: %u COUNT: %d\n", u, c);
  
  return c;
}

void Inheritance_Matrix::Print_Inheritance_Matrix()
{
  for(int i=0; i<famSize; i++)
    {
      Person *p;
      p = ped->persons[fam->path[i]];
      printf("%s ",p->pid.c_str());
    }
  
  for(int i=0; i<nInherit_vector; i++)
    { 
      cout << bitset<20>(inherit_vector[i]).to_string()<<":\t";
      for(int j=0; j<famSize*2; j++)
	printf("%d \t", inherit_matrix[i][j]);
      printf("\n");
    }
}

void Inheritance_Matrix::Print_Founder_Node_Indicator()
{
  for(int i=0; i<pow(4, nFounders); i++)
    {
      for(int j=0; j<two_f; j++)
	printf("%d ",  founder_node_indicator[i][j]);
      printf("\n");
    }
}

void Inheritance_Matrix::Print_Inheritance_GenoIdx()
{
  for(int i=0; i<famSize; i++)
    {
      Person *p;
      p = ped->persons[fam->path[i]];
      //printf("%s ",p->pid.c_str());
    }
  printf("\n");
  for(int i=0; i<nInherit_vector; i++)
    { 
      cout << bitset<20>(inherit_vector[i]).to_string()<<":\n";
      for(int j=0; j<famSize; j++) 
	{
	  for(int k=0; k<four_pow_f; k++)
	    printf("%d \t", geno_index[i][k][j]);
	  printf("\n");
	}
    }
}

