#ifndef __PEDMARKERDATA_H__
#define __PEDMARKERDATA_H__

#include "StringArray.h"
#include "Pedigree.h"
#include "MapFunction.h"
#include <stdint.h>

#include <cassert>
#include <string>
#include <set>
#include <map>
#include <vector>
#include <algorithm>
#include <math.h>

#define INIT_SIZE 1000000
#define GLLIM 255 // GLLIM should be <1024 in current version

using namespace std;

class VarInfo
{
 public:
  std::string chr;
  int pos;
  std::string id;
  std::string ref;
  std::string alt;
  float qual;
  std::string filter;
  std::string info;
  std::string format;
  int ac;
  int dp;
  float freq;
  float map_dist;

 public:
  VarInfo(){
   chr="";
   pos=-1;
   id = ".";
   ref="";
   alt="";
   qual = 0.0;
   filter = ".";
   map_dist=-1;
  }

 VarInfo& operator= (const VarInfo &vSource)
 {
   chr = vSource.chr;
   pos = vSource.pos;
   ref = vSource.ref;
   alt = vSource.alt;
   map_dist = vSource.map_dist;
   return *this;
 }
};

class MapInfo
{
 public:
  std::vector<std::string> chr;
  std::vector<std::string> marker_name;
  std::vector<int> pos;
  std::vector<float> distance;
  std::vector<float> recomb;
};

class SeqMarkerInfo
{
 public:
  float geno_lk[3];
  uint32_t geno_pl[3];
  int dp;
  bool missing_dp;
  bool missing_pl;

 public:
  SeqMarkerInfo(){
   geno_lk[0]=geno_lk[1]=geno_lk[2]=1.0;
   geno_pl[0]=geno_pl[1]=geno_pl[2]=0;
   dp = 0;
   missing_dp = false;
   missing_pl = false;
  }

 SeqMarkerInfo& operator= (const SeqMarkerInfo &mSource)
 {
  geno_lk[0] = mSource.geno_lk[0];
  geno_lk[1] = mSource.geno_lk[1];
  geno_lk[2] = mSource.geno_lk[2];

  geno_pl[0] = mSource.geno_pl[0];
  geno_pl[1] = mSource.geno_pl[1];
  geno_pl[2] = mSource.geno_pl[2];

  dp = mSource.dp;
  missing_dp = mSource.missing_dp;
  missing_pl = mSource.missing_pl;
  return *this;
 }
};

class PedMarkerData
{
 public:
  int size;
  int count;
  int nFam;
  int *famSizes;
  int GL_idx;
  int PL_idx;
  Pedigree *ped;
  SeqMarkerInfo ***markers;
  float PL2LK_table[GLLIM+1];
  std::string chr; int start; int end;
  std::vector<VarInfo> var;
  std::vector<std::string> pids;
  std::vector<std::string> pids_with_data;
  std::map<std::string, std::pair<int, int> > pid2traverse;
  std::map<std::pair<int, int>, std::string > traverse2pid;
  std::map<std::pair<int, int>, int > fam_pid_index_pair_with_data;
  MapInfo marker_map;
  float submap;

 public:
  PedMarkerData();
  ~PedMarkerData();
  PedMarkerData(Pedigree *);
  void Initialize();
  void SetPedigree(Pedigree*);
  void SetInputRange(std::string, int, int);
  void MapPID2Traverse();
  void SetSubmap(float);
  void LoadMarkerVCF(String);
  void LoadMarkerVCF(FILE *);
  void LoadMapFile(FILE *);
  void LoadMapFile(String&);
  void AllocateMarkerMemory();
  void AllocateMarkerMemory(SeqMarkerInfo***&);
  void ReAllocateMarkerMemory();
  void DeleteMarkerMemory();
  void CopyMarkerMemory(SeqMarkerInfo***);
  void AddMarkerInfo();
  void PrintGL();
  float PL2LK(int);
};

#endif
